/* 
 * ControlCGA.h 
 * 
 * Proposito: Controlar la simulacion del video CGA de una PC. 
 * 
 * Esta simulacionn trabaja con 3 modos de video: 
 *    MODO        DECLARACION EN   RESOLUCION 
 *                 LA INT 0x10 
 *    Texto           0x07 	   80x25 
 *    4ColorGraphic   0x04	  320x200   
 *    2ColorGraphic   0x06        640x200   
 * 
 * Si el programa selecciona un modo distinto a los declarados, por 
 * omision se coloca en modo texto.
 * El modo texto trabaja con 4 paginas de texto, cada pagina se simula 
 * con un objeto ScrollView. 
 * 
 * Amilcar Meneses Viveros 
 * [3-febrero-2001]
 * 
 */ 
#import <appkit/appkit.h>
#import "CGAView.h"

/* 
 * Modos de video. 
 * Texto: 80x25 
 * Graf1: 320x200  (4 colores, 2 a la vez)  0xB000:0x8000
 * Graf2: 640x200  (2 colores, 2 a la vez)  0xB000:0x8000
 */ 

/* Modos definidos en la INT 0x10 */
#define MODO_TEXTO	0x07
#define MODO_GRAF1	0x04
#define MODO_GRAF2 	0x06

#define VIDEO_SET_DISPLAY_MODE	0x00
#define VIDEO_SET_CURSOR_TYPE	0x01		/* No imaplantada */ 
#define VIDEO_SET_CSR_POSITION	0x02
#define VIDEO_GET_CSR_POSITION	0x03
#define VIDEO_GET_PEN_POSITION  0x04 		/* No implantada  */ 
#define VIDEO_SET_DISPLAY_PAGE	0x05
#define VIDEO_SCROLL_WIN_UP	0x06		/* No implantada  */ 
#define VIDEO_SCROLL_WIN_DOWN	0x07	 	/* No implantada  */ 
#define VIDEO_READ_ATRB_CHAR	0x08		
#define VIDEO_WRITE_ATRB_CHAR	0x09
#define VIDEO_WRITE_CHAR_CURSOR	0x0a
#define VIDEO_SET_PALETTE	0x0b		/* No implantada  */
#define VIDEO_WRITE_PIXEL	0x0c
#define VIDEO_READ_PIXEL	0x0d
#define VIDEO_TTY_OUT		0x0e
#define VIDEO_GET_DISPLAY_MODE	0x0f

#define MASK8(x)	((x) & 0xff)
#define MASK16(x)	((x) & 0xffff)
#define HIGH(x)		MASK8((unsigned long)(x) >> 8)
#define LOW(x)		MASK8((unsigned long)(x))
#define WORD(x)		MASK16((unsigned long)(x))
#define SETHIGH(x,y) 	(*(x) = (*(x) & ~0xff00) | ((MASK8(y))<<8))
#define SETLOW(x,y) 	(*(x) = (*(x) & ~0xff) | (MASK8(y)))
#define SETWORD(x,y)	(*(x) = (*(x) & ~0xffff) | (MASK16(y)))

typedef struct _VideoTexto 
{
   unsigned row; 
   unsigned column; 
   char memVideo[25][81]; 
} VideoTexto; 


@interface ControlCGA:Object
{
    id	multiView;
    id	pag1;			// View de pagina 1
    id	pag2;			// View de pagina 2
    id	pag3;			// View de pagina 3 
    id	pag4;			// View de pagina 4 
    id  apag[4]; 

    id  text1; 			// Pagina modo texto 1
    id  text2; 			// Pagina modo texto 2
    id  text3; 			// Pagina modo texto 3 
    id  text4; 			// Pagina modo texto 4 
    id  text;
    id  atext[4]; 

    id  vgraf; 			// View de modo grafico 
    id  graf; 			// Pagina grafica 

    unsigned modo;  
    int pagina; 
    VideoTexto texto[4]; 
    int bandVideo; 
}

- awakeFromNib ;

- setPag: sender;
- setPagToView: theView; 

- setActivePage; 
- setActiveGraph;

- display; 
- do_int10_bios; 
- clearBandera; 
@end
