#import "VirtualMode.h"
#import <sys/dir.h>
#import <libc.h> 

#define LAP  100
#define LAC  100

@implementation VirtualMode

static port_t guiPort;  	// Hilo principal,recibe y muestra mensajes
extern port_t port_exc;		// Hilo monitor, recibe mensaje de finalizacion

int           bandera;        // global variable of current tick count 
mutex_t       ExcLock;		// mutual exclusion to protect bandera 
condition_t   ExcUpdated;      // signaled when bandera is increased 

static char   **lista; 		// lista de excepciones 
static char   *ea[LAP]; 
lstexcep      *rle; 		// lista de estructura de excepcion 

extern BOOL stopException; 

void init_msg(simpleMsg *m);

// extern      void tarea_monitor(); 

static void msgHandler(msg, self)
simpleMsg *msg;
id self;
{
    [self draw:msg];
}

+ initialize 
{
    if (self == [VirtualMode class]) {
	bandera = 0;
	ExcLock = mutex_alloc();
	ExcUpdated = condition_alloc();
    }

    return self;
}

- init
{
    int i;

   [super init];
   port_allocate(task_self(), &guiPort);
   DPSAddPort(guiPort, msgHandler, sizeof(simpleMsg), self, 
               NX_MODALRESPTHRESHOLD +1);

   [NXApp loadNibSection:"VirtualMode.nib" owner:self];
   *directorio = '\0';
   [[NXBundle mainBundle] getPath:directorio 
                          forResource:"excepciones.txt" 
                          ofType:NULL];
   for (i=0; i<LAP; i++) ea[i] = NULL; 
   rle = NULL; 
   lista = NULL;
   monitor = memoria = nil; 
 
   offset = 0.0;
   tarea_monitor();
   stopException=YES; 

   return self;
}

- free
{
   [self endExecute]; 
   [self endVMTask]; 
   return [super free]; 
}

- setOffset:(float)valor
{
   offset = valor;
   return self;
}

- setStopException:sender 
{
   stopException = (stopException==YES)?NO:YES;
   return self;  
}

- execute:sender
{
   if (monitor == nil) { 
      monitor = [[ThreadEngine alloc] init]; 
      [monitor fork:(cthread_fn_t)maneja_hiloV86 arg: (any_t) 0]; 

      memoria = [[ThreadEngine alloc] init]; 
      [memoria fork:(cthread_fn_t)memoriaV86 arg:(any_t) 0]; 
   } else { 
      /* codigo para desactivar las ejecuciones del hilo virtual, 
       * del hilo monitor, y del hilo de memoria 
       */
      [self endExecute];
   }

   return self;
}

- continue:sender 
{
  mutex_lock(ExcLock);
  bandera = 1;
  mutex_unlock(ExcLock);
  condition_signal(ExcUpdated);
  return self; 
}

- show:sender
{
    NXRect frame; 

    [myWindow getFrame:&frame];
    NX_X(&frame) += offset; 
    NX_Y(&frame) -= offset; 
    [myWindow placeWindowAndDisplay: &frame];
    [myWindow makeKeyAndOrderFront:self];
    return self;
}

- carga:(char *)nombrearchivo
{
   id cargaArchivo; 

   cargaArchivo = [[ThreadEngine alloc] init]; 
   [cargaArchivo fork:(cthread_fn_t)carga arg: (any_t)nombrearchivo]; 
   [buttonExe setEnabled:YES]; 
   [threadState setStringValue: nombrearchivo]; 
   [myWindow setTitle: nombrearchivo];
   [myWindow makeKeyAndOrderFront:nil];

   return self; 
}

- muestraPrograma:(char *)cadena
{
    int l;
    Text *texto; 

    texto = [programa docView];  
    l = [texto textLength]; 
    [texto setSel:l :l]; 
    [texto replaceSel: cadena]; 
    [texto scrollSelToVisible]; 
    [texto display]; 
    return self; 
} 

 
- endExecute
{
   port_t         old_exc_port, clear_port;
   msg_return_t ret; 

   ret = task_get_exception_port(task_self(), &old_exc_port);
   if (ret != KERN_SUCCESS) {
      //mach_error("task_get_exception_port", ret);
       exit(1);
   }

   clear_port = thread_reply();
   ret = exception_raise(port_exc, clear_port,
        thread_self(), task_self(), EXC_SOFTWARE, END_RUN_VM, 0);

// if (ret != KERN_SUCCESS) 
//     mach_error("catch_exception_raise didn't handle exception",
//                ret);
// else printf("Successfully called exception_raise!!!!!!!!!!!\n");
 
//   printf("Sigue\n");
   ret = task_set_exception_port(task_self(),  old_exc_port);
//   printf("....y sigue\n");
   if (ret != KERN_SUCCESS) {
      mach_error("task_set_exception_port", ret);
      exit(1); 
   }
   printf("Termina monitor y memoria\n");
   if (monitor!=nil) { 
      [monitor abort]; 
      [monitor free]; 
      monitor = nil; 
   }
   if (memoria!=nil) { 
      [memoria abort]; 
      [memoria free]; 
      memoria = nil; 
   }
   printf("Terminaron bien?\n");
   return self; 
}

- endVMTask
{
   port_t       old_exc_port, clear_port;
   msg_return_t ret; 

   ret = task_get_exception_port(task_self(), &old_exc_port);
   if (ret != KERN_SUCCESS) {
       mach_error("task_get_exception_port", ret);
       exit(1);
   }

   clear_port = thread_reply();
   ret = exception_raise(port_exc, clear_port,
        thread_self(), task_self(), EXC_SOFTWARE, END_TASK_VM, 0);

// if (ret != KERN_SUCCESS) 
//     mach_error("catch_exception_raise didn't handle exception",
//                ret);
// else printf("Successfully called exception_raise!!!!!!!!!!!\n");
 
//   printf("Sigue\n");
   ret = task_set_exception_port(task_self(),  old_exc_port);
//   printf("....y sigue (fin tarea)\n");
   if (ret != KERN_SUCCESS) {
      mach_error("task_set_exception_port", ret);
      exit(1); 
   }
   return self; 
}

/* Metodos y funciones para comunicacion con el puerto asociado a DPS */ 
void init_msg(simpleMsg *m) 
{
//   printf("init_msg, guiPort=%d\n", guiPort);
   m->h.msg_remote_port = guiPort; //set port to send drawing to
   m->h.msg_simple = TRUE; 	       // no ports or out-of-line data
   m->h.msg_size = sizeof(simpleMsg);
   m->h.msg_local_port = PORT_NULL;
   m->h.msg_type = MSG_TYPE_NORMAL;
   m->t.msg_type_name = MSG_TYPE_INTEGER_32;
   m->t.msg_type_size = 32;
   m->t.msg_type_number = MAXDATA;
   m->t.msg_type_inline = TRUE;	        // no out-of-line data will be passed
   m->t.msg_type_longform = FALSE;
   m->t.msg_type_deallocate = FALSE;
}

- draw:(simpleMsg *)msg
{
    int l, i;
    Text *texto; 
    lstexcep *ce; 
    u_char code; 
    id cga;

    switch(msg->h.msg_id) {
    case STATE_MSG: [threadState setStringValue:"Estado del hilo"]; 
                    break; 	
                           
    case EXCEP_MSG: code = (int)msg->body[0];
                //  printf("Se muestra la excepcion %d 0x%4.4x\n",code, code);
                    ce = busca_codigo(rle, (int)code); 
                    if (ce == NULL) { 
                        i=0; 
                      printf("No encuentro codigo!\n");
                    }
                    else { 
                        i  = ce->i;
                        printf("Encuentro codigo en indice %d!\n", i);
                    } 
                    [[typeException matrixInColumn:0] selectCellAt: i:0];
                    [typeException sendAction];  
                    break; 
    case EXECT_MSG: // [execute ]  
                    break; 
    case FNSHD_MSG: break; 
    case CLEAR_MSG: 
                    texto = [execute docView];  
//                    [texto clear:texto]; 
                    [texto selectAll:texto]; 
                    [texto delete: texto]; 
//                    [texto setSel:0 :0]; 
//                    [texto replaceSel: "\0"]; 
                    [texto scrollSelToVisible]; 
                    [texto display]; 
//                    l = 0; 
                    [threadState setStringValue: "BORRE!"]; 
                    break; 
    case MEMOR_MSG: 
                    texto = [execute docView];  
                    l = [texto textLength]; 
                    [texto setSel:l :l]; 
                    [texto replaceSel: msg->body]; 
                    [texto scrollSelToVisible]; 
                    [texto display]; 

                    [threadState setStringValue: "YES!"]; 
                    break; 
    case PROGR_MSG: 
                    texto = [programa docView];  
                    l = [texto textLength]; 
                    [texto setSel:l :l]; 
                    [texto replaceSel: msg->body]; 
                    [texto scrollSelToVisible]; 
                    [texto display]; 
                    break; 
    case NLOAD_MSG: 
                    NXRunAlertPanel(0, 
                    "No se puede abrir el archivo", 0, 0, 0); 
                    break; 
    case SIMVI_MSG: cga = [NXApp controlCga]; 
                    [cga do_int10_bios]; 
                    [cga clearBandera]; 
                    break; 
    }
    return self;
}


/* Metodos del delegado Browser */  

/* Regresa el indice del caracter ',' en la cadena w */ 
/* Amilcar Meneses Viveros 			     */ 
/* [26-Julio-1998]				     */ 
int buscar_car(w)
char *w; 
{
   int i;
   for(i=0; i<LAC; i++) 
     if (w[i] == ',') return i; 
   
   return -1; 
}

/* Inserta un nuevo elemento a una lista de lstexcep  */ 
/* Esta lista esta enumerada en orde asendente, en el */ 
/* rango de 0 a n-1.				      */ 
/* Amilcar Meneses Viveros			      */ 
/* [26-Julio-1998]				      */ 
lstexcep *inserta_lista_estructura(l,w,k,j)
lstexcep *l; 
char     *w;
int       k; 
int      *j; 
{
   int i;
   if(l==NULL) { 
      l = (lstexcep *)malloc(sizeof(lstexcep)); 
      l->i = k; 
      i = buscar_car(w); 
      l->e    = (char *)malloc((i+1)*sizeof(char)); 
      strncpy(l->e, w, i); 
      l->e[i] = '\0'; 
      l->next = NULL; 
      l->code  = atoi(w+i+1); 
      *j = k; 
      return l; 
   } 
   else l->next = inserta_lista_estructura(l->next, w, k+1, j);
   return l; 
}

lstexcep *busca_indice(l, i)
lstexcep *l; 
int       i;
{
   if (l==NULL) return NULL; 
   else {
//       printf("(%d), %s = %d\n", i, l->e, l->code); 
       if (l->i == i) return l; 
       else return busca_indice(l->next, i); 
   } 
}

lstexcep *busca_codigo(l, c)
lstexcep *l; 
int       c;
{ 
   if (l==NULL) return NULL; 
   else{ 
//      printf("(%d), %s = %d\n", c, l->e, l->code); 
      if (l->code == c) return l; 
      else return busca_codigo(l->next, c); 
   }
}

- (int)browser:sender fillMatrix:matrix inColumn:(int) column 
{
    int i, j; 
    char line[LAC];
    FILE *fp; 
    lstexcep *ce; 

    i = j = 0;
    if((fp=fopen(directorio, "r"))==NULL) { 
       sprintf(line, "No puedo abrir el archivo (%s)", directorio); 
       [threadState setStringValue:line]; 
       return 0;
    } 
    sprintf(line, "Abri el archivo (%s)", directorio); 
    [threadState setStringValue: line];  
    while( fgets(line,LAC,fp)!=NULL )
    {
       rle = inserta_lista_estructura(rle, line, 0, &j); 
       ce = busca_indice(rle, j); 
       ea[i] = ce->e; 
       // printf("code[%d] = %d\n", i,ce->code); 
       i++; 
    }
    fclose(fp); 
    lista = ea; 
    return i; 
} 

- browser:sender loadCell:cell atRow:(int)row inColumn:(int) column 
{
   if(lista) { 
     [cell setStringValueNoCopy: lista[row]]; 
     [cell setLeaf:YES]; 
   }

   return self; 
}

@end
