
#import "tv86.h" 
#import "dasm.h"
#import "task_vm.h"
//#import "sim.h"
#import <objc/objc.h>

#define BIOS_VIDEO	0x10

#define MEM_INICIAL	0x000000
#define MEGABYTE	0x100000 
#define NBYTES		0x10

#define LOAD_SEGMENT	0x0200
#define LOAD_OFFSET     0x0000

#define I386_NOP	0x90
#define I386_INT	0xcd
#define I386_HLT	0xf4
#define I386_CLI	0xfa
#define I386_STI	0xfb
#define I386_PUSHF	0x9c
#define I386_POPF	0x9d
#define I386_LOCK	0xf0
#define I386_IRET	0xcf
#define I386_IN1	0xe4
#define I386_IN2	0xe5
#define I386_IN3	0xec
#define I386_IN4	0xed
#define I386_OUT1	0xe6
#define I386_OUT2	0xe7
#define I386_OUT3	0xee
#define I386_OUT4	0xef
#define I386_INS1	0x6c
#define I386_INS2	0x6d
#define I386_OUTS1	0x6e
#define I386_OUTS2	0x6f
#define I386_REPZ	0xf3
#define I386_OP_SIZE	0x66		/* Operacion del i386 */ 

task_t        new_t; 
thread_t      vm_thread; 
port_t        port_exc; 

char 	      *addr_cache; 			/* Pagina cache */

BOOL stopException; 

extern void init_msg(simpleMsg *m); 		/* Funcion de inicializacion 
 						 * de mensaje con el objeto 
                                                 * VirtualMode 
                                                 */ 

extern void desasembler(char *buff, int n);	/* Funcion que llama al 
   						 * desensamblador 
 						 */ 

extern void do_int10_bios(i386_thread_state_t *state);	
                                                /* Funcion que llama a la 
						 * simulacion de la int 0x10 
						 * del BIOS 
						 */ 

extern int           bandera;
extern mutex_t       ExcLock;  
extern condition_t   ExcUpdated;  

void informa_tarea(task_t *t);

/*
 * Funcion para crear el ambiete de una tarea virtual,      
 * esto es, debe crea una tarea hija, asignale el 1er. 
 * megabyte de memoria y quitarle las protecciones de lectura 
 * escritura y ejecucion. 
 * Amilcar Meneses Viveros 
 * [6-MAR-1998]
 */ 
void tarea_virtual(task_t *t)
{
    char         *addr; 
//    vm_size_t	  size_ret; 
    kern_return_t rtn; 

    if ( (rtn = task_create(task_self(), FALSE, t)) != KERN_SUCCESS) { 
/*        mach_error("task_create fallo ",rtn); */
        printf("Error al crear una nueva tarea!!!!\n");
        exit(-1);
    }

    addr = (char *)0x00; 
    if((rtn = vm_allocate(*t, (vm_address_t *) &addr, MEGABYTE, FALSE)) 
          != KERN_SUCCESS) { 
        mach_error("vm_allocate fallo",rtn);
        printf("Error en vm_allocate.\n"); 
        exit(-1); 
    } 

//    informa_tarea(t); 
/*
    if(addr!= 0x00) { 
       printf("Se cambio la direccion 0x00 por %8.8x\n", (u_int) addr); 
       exit(-1); 
    }
 */ 
 
    addr = (char *) 0x40; 
    if ((rtn=vm_protect(*t, (vm_address_t) addr, MEGABYTE, 
            FALSE, 0x07)) != KERN_SUCCESS) {
       mach_error("vm_protect fallo ",rtn);  
       // printf("Error en vm_protect\n"); 
       exit(-1); 
    }
}


/* ===================================
     Funciones Para Mostrar Memoria 
   =================================== */

/* 
 * Funcion que muestra el estado de unbloque de memoria, esto es, 
 * despliega su tamanho, sus protecciones y los atributos de memoria 
 * para las tareas hijas 
 * Amilcar Meneses Viveros. 
 * [12-Marzo-1998] 
 */ 
void estado_mem(task_t *t, vm_address_t addr)
{
//  char          *data;
  kern_return_t r;
  vm_size_t     size;
  vm_prot_t     protection, max_protection;
  vm_inherit_t  inheritance;
  boolean_t     shared;
  port_t        object_name;
  vm_offset_t   offset;
  vm_address_t  address; 

  address = addr; 
  r = vm_region(*t, &address, &size, &protection, 
      &max_protection, &inheritance, &shared, &object_name, &offset);

  if (r != KERN_SUCCESS) {
      mach_error("Error al llamar a vm_region", r);
      if (r==KERN_NO_SPACE) 
         mach_error("La tarea no tiene la direccion solicitada"); 
      exit(-1); 
  }
  else {
      printf("La herencia es:   ");
      switch (inheritance) {
      case VM_INHERIT_SHARE:  
           printf("Compartida con hijo\n");
           break;
      case VM_INHERIT_COPY:  
           printf("Copiada al hijo\n");
           break;
      case VM_INHERIT_NONE:  
           printf("Absent from child\n");
           break;
      case VM_INHERIT_DONATE_COPY:
           printf("Copy and delete\n");
           break;
    }
    printf("El tamanho es:    %d (%8.8x)\n", size, size); 
    printf("La proteccion es: "); 
    switch(protection){
    case VM_PROT_NONE: 
         printf("(VM_PROT_NONE) no permite acceso\n"); 
         break; 
    case VM_PROT_READ:
         printf("(VM_PROT_READ) permite acceso de lectura\n");
         break; 
    case VM_PROT_WRITE:
         printf("(VMPROT_WRITE) permite accesos de lectura, escritura \n"); 
         printf("\t\t\t y ejecucion\n"); break;
    case VM_PROT_EXECUTE:
         printf("(VM_PROT_EXECUTE) permite acceso de ejecucion\n");
    default: 
         printf("No existe proteccion definida (%x)!!!!!\n", protection); 
         break; 
    }
  }
}

/* 
 * Funcion que muestra el estado de una direccion de memoria 
 * al estilo 8086 
 * Amilcar Meneses Viveros. 
 * [14-Marzo-1998] 
 */  
void estado_mem8086(task_t *t, segment_8086 s, offset_8086 o)
{
  vm_address_t addr; 
  addr = (vm_address_t) Address_8086(s,o); 
  printf("\n El estado de memoria %8.8x:%8.8x es:\n", s, o); 
  estado_mem(t, addr); 
}

/* 
 * Funcion que muestra nbytes de la memoria de la tartea virtual (*t) 
 * a partir de la direccion addr. 
 * Amilcar Meneses Viveros 
 * [9-MARZO-1998] 
 */ 
void muestra_memoria(task_t *t, vm_address_t addr, u_int nbytes)
{
   char  *datos, *addr_i, *addr_j; 
   int   nb, i, j, k, nl; 
   kern_return_t error; 

   j = i = nl = 0; 
   if (error=vm_read(*t, (vm_address_t) addr, vm_page_size, 
      (pointer_t *)&datos, &nb) != KERN_SUCCESS ) { 
      mach_error("vm_read fallo en muestra memoria", error); 
    /*  printf("Error en la lectura del bloque de memoria \n");  */
      exit(-1); 
   } 

   addr_i =  datos;  
   for (j=0; j < nbytes; j++) {
       addr_j = addr_i; 
       if ( nl >0 && ((nl++ % 10)==0) ) getchar(); 
       printf("\n[%4.4x] ", addr+j);
       for (k=0; k<NBYTES; k++) {
           if (k==8) printf("- "); 
           if (i < nbytes) printf("%2.2x ", (unsigned char)*addr_j++); 
           else            printf("__"); 
       }
  
       addr_j = addr_i; 
       for (k=0; k < NBYTES; k++) {
           i++; addr_j++;  
           if (k == 7) printf("-");
           if (*(addr_j)>=' ' && *(addr_j)<='~') 
              printf("%c", *(addr_j)); 
           else  printf("."); 
       }
          addr_i += NBYTES; 
          j += NBYTES; 
      }

   if ( (error=vm_deallocate(task_self(), (vm_address_t) datos, vm_page_size) ) 
         != KERN_SUCCESS ) {
      mach_error("Fallo vm_deallocate ", error); 
      printf("Salgo de la funcion muestra_memoria\n"); 
      exit(-1); 
   }
   printf("\n"); 
}

/* 
 * Muestra n bytes de memoria de la tarea virtual (*t) a partir de 
 * de la direccion segment:offset (s:o) al estilo 8086. 
 * Amilcar Meneses Viveros. 
 * [14-Marzo-1998] 
 */ 
void muestra_memoria8086(task_t *t, segment_8086 s, offset_8086 o, u_int n) 
{
   vm_address_t addr; 
   addr = (vm_address_t) Address_8086(s,o); 
   muestra_memoria(t, addr, n); 
}

/* Funcion de prueba de escritura a memoria */ 
void llena(task_t *t, vm_address_t *dat)
{
   int i; 
   u_char c; 
   kern_return_t rtn; 
   char *dato, *addr;  

   /* Crea 1 megabyte que sera pasado a la tarea hija */ 
   if ( (rtn =  vm_allocate(task_self(), (vm_address_t *) &dato, 
          MEGABYTE, TRUE) ) != KERN_SUCCESS ) { 
      mach_error("vm_allocate fallo ", rtn ); 
   /*   printf("Salgo en funcion llena \n"); */
      exit(-1); 
   }

   for (i=0, c='0'; i<MEGABYTE; i++) { 
       if ( (i%8192) == 0 ) c++; 
           dato[i] = '\0'; /* dato[i] = c; */
   }     

   addr = (char *) 0x00; 
   if ( (rtn = vm_write(*t , (vm_address_t) addr, (pointer_t) dato, 
        MEGABYTE) )!= KERN_SUCCESS ) {
      mach_error("Fallo vm_write", rtn); 
   /*   printf("Salgo en funcion error \n");  */
      exit(-1); 
   }

   *dat = (vm_address_t) addr; 
   if ( (rtn = vm_deallocate(task_self(), (vm_address_t) dato, MEGABYTE) )!= 
         KERN_SUCCESS ) { 
       mach_error("Fallo vm_deallocate ", rtn); 
    /*   printf("Salgo en funcion llena ");  */ 
       exit(-1); 
   }
}


/* 
 * Funcion que despliega la informacion de la tarea (*t) 
 * Amilcar Meneses Viveros 
 * Hugo Garcia Monroy 
 * [16-Marzo-1998] 
 */ 
void informa_tarea(task_t *t)
{ 
   kern_return_t             rtn;
   struct task_basic_info    info;
   unsigned int              info_count=TASK_BASIC_INFO_COUNT;

   
   if ((rtn=task_info(*t, TASK_BASIC_INFO, 
       (task_info_t)&info, &info_count))!=KERN_SUCCESS) { 
       mach_error("fallo task_info()", rtn); 
       printf("Salgo de la funcion informa_tarea\n"); 
       exit(-1); 
   } 
   else { 
       printf("suspend_count    :%d\n", info.suspend_count);
       printf("Prioridad Base   :%d\n", info.base_priority);
       printf("Tamanho Virtual  :%d 0x%8.8x\n", 
             info.virtual_size/vm_page_size, info.virtual_size); 
       printf("Tamanho Residente:%d 0x%8.8x\n", 
             info.resident_size/vm_page_size, info.resident_size); 
       printf("Tiempo de usuario:%ld,%ld\n", 
            (long int) info.user_time.seconds, 
            (long int) info.user_time.microseconds); 
       printf("Tiempo de sistema:%ld,%ld\n", 
            (long int) info.system_time.seconds, 
            (long int) info.system_time.microseconds); 
   } 
}

/* 
 * Funcion que carga un archivo al inicio de una pagina de 
 * memoria virtual. 
 * Amilcar Meneses Viveros. 
 * [16-Marzo-1998]
 */ 
int carga_archivo(task_t t, char *name, vm_address_t addr, int *num)
{
   int i, fd, n; 
   kern_return_t rtn; 
   char *dato, *addr_i;  

   simpleMsg mensaje; 
   init_msg(&mensaje); 
   mensaje.h.msg_id = NOT_LOAD_MSG; 

   /* Crea 1 megabyte que sera pasado a la tarea hija */ 
   if ( (rtn =  vm_allocate(task_self(), (vm_address_t *) &dato, 
          PAGINA_VIRTUAL, TRUE) ) != KERN_SUCCESS ) { 
      mach_error("vm_allocate fallo ", rtn ); 
      printf("Salgo en funcion llena \n"); 
      msg_send((msg_header_t *)&mensaje, MSG_OPTION_NONE,0);
      return -1; 
   } 

   if ( (fd=open(name, O_RDONLY)) < 0 ) { 
      printf("No puedo abrir el archivo %s\n", name); 
      printf("Salgo de la funcion carga_archivo\n"); 
      msg_send((msg_header_t *)&mensaje, MSG_OPTION_NONE,0);
      return -1; 
   } 

   *num = 0; 
   addr_i = (char *) addr; 
   while ( (n=read(fd, dato, PAGINA_VIRTUAL)) > 0 ) { 
     for (i=n+1; i<PAGINA_VIRTUAL; i++)  dato[i] = '\0'; 

     if ( (rtn = vm_write(t , (vm_address_t) addr_i, (pointer_t) dato, 
          PAGINA_VIRTUAL) )!= KERN_SUCCESS ) {
        mach_error("Fallo vm_write", rtn); 
        printf("Salgo en funcion carga_archivo 0x%8.8x\n", addr_i);  
        msg_send((msg_header_t *)&mensaje, MSG_OPTION_NONE,0);
        return -1; 
     }
     desasembler(dato, n); 
     addr_i += vm_page_size; 
     *num += n; 
   }

   if ((rtn=vm_deallocate(task_self(),(vm_address_t) dato, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
       mach_error("Fallo vm_deallocate ", rtn); 
       printf("Salgo en funcion llena "); 
       msg_send((msg_header_t *)&mensaje, MSG_OPTION_NONE,0);
       return -1; 
   }
   printf("Carge la tarea\n"); 
   return 1; 
}


/* 
 * Funcion que crea un hilo virtual en la tarea (*vmt) 
 * Amilcar Meneses Viveros 
 * [17-febrero-1998]
 */ 
void thread_V86(task_t t, thread_t *vmt, vm_address_t entry_point, port_t *pe)
{
  int                 vm_state_count;    /* Longitud de estr. regs.  */
  i386_thread_state_t vm_thread_state;   /* estructura con registros */
  kern_return_t       error;             /* tipo de error            */

  /* Tamanho de la estructura hilo entre la longitud del int */
  vm_state_count = i386_THREAD_STATE_COUNT; 

  /* Se crea el hilo */ 
  if ((error =  thread_create(t, vmt)) != KERN_SUCCESS) { 
      mach_error("fallo thread_create ",error);
      printf("Salgo de la funcion thread_V86\n"); 
      exit(-1); 
  }

  /* Crea el puerto de exepciones */ 
  if ((error=port_allocate (task_self(), pe)) != KERN_SUCCESS) {
    mach_error("Fallo port_allocate en thread_V86", error); 
    exit(1);
  }

  /* asigna el puerto de exepciones */ 
  thread_set_special_port(*vmt, THREAD_EXCEPTION_PORT, *pe); 
/*  printf("<<<<< 0x%8.8x, 0x%8.8x -> %d >>>>>\n", 
          (u_int) &port_exc, (u_int) pe, *pe);    */ 


  /* Obtiene el estado del hilo y lo inicializa */ 
  if ((error=thread_get_state(*vmt, i386_THREAD_STATE, 
     (thread_state_t)(&vm_thread_state), &vm_state_count ))!=KERN_SUCCESS) { 
     mach_error("Fallo el llamado a thread_get_state ", error); 
     printf("Salgo de la funcion thread_V86\n"); 
     exit(-1); 
  }

  vm_thread_state.eflags  = (vm_thread_state.eflags | EFL_VM); 
  vm_thread_state.eip = entry_point; 
  vm_thread_state.cs  = 0x00; 
  vm_thread_state.ds  = 0x00; 
  vm_thread_state.ss  = 0x00; 
  vm_thread_state.gs  = 0x00; 
  vm_thread_state.es  = 0x00; 
  vm_thread_state.ebp = 0x00; 
  vm_thread_state.eax = 0x00; 
  vm_thread_state.ebx = 0x00; 
  vm_thread_state.ecx = 0x00; 
  vm_thread_state.edx = 0x00; 
  vm_thread_state.edi = 0x00; 
  vm_thread_state.esi = 0x00; 
  vm_thread_state.esp = 0x300;  


  /* Carga el estado del hilo */ 
  if((error  = thread_set_state(*vmt, i386_THREAD_STATE, 
     (thread_state_t)(&vm_thread_state), vm_state_count))!=KERN_SUCCESS) { 
      mach_error("Fallo el llamado a thread_set_state", error); 
      printf("Salgo de la funcion thread_V86\n"); 
      exit(-1); 
  }
}

void informa_hilo(thread_t t)
{
  kern_return_t             error;
  struct thread_basic_info  info;
  unsigned int              info_count=THREAD_BASIC_INFO_COUNT;

  error=thread_info(t, THREAD_BASIC_INFO, (thread_info_t)&info, &info_count);
  if (error!=KERN_SUCCESS)
      mach_error("Error calling thread_info()", error);
  else {
      printf("Tiempo de usuario %d seg, %d microseg\n", 
          info.user_time.seconds, info.user_time.microseconds);
      printf("Tiempo de sistema %d seg, %d microseg\n", 
          info.system_time.seconds, info.system_time.microseconds);
      printf("Uso de CPU  : %d\n", info.cpu_usage); 
      printf("Estado de ejecucion %d\n", info.run_state); 
      printf("Contador de suspencion %d\n", info.suspend_count);
      printf("Banderas     : 0x%8.8x\n", info.flags);
      printf("Tiempo de desactivacion %d\n", info.sleep_time); 
   }

}

/*************************************************
 * Funciones de mapeo (lectura-escritura)        *
 *************************************************/ 
/* 
 * Funcion que duplica la pagina que contiene la direccion vmeip a una pagina 
 * local y mapea la direccion vmeip a una direccion eip correspondiente a la 
 * pagina duplicada. 
 * Lectura 
 * Amilcar Meneses Viveros 
 * [9-JULIO-2000]
 */ 
kern_return_t map_in(task_t *t, char *vmeip, 
                     vm_address_t *eip, vm_address_t *addr)
{
   int nb; 
   char *addr_in, *temp; 
   kern_return_t error;

   addr_in = (char *)trunc_page(vmeip);
   if (error=vm_read(*t, (vm_address_t) addr_in, vm_page_size, 
      (pointer_t *)addr, &nb) != KERN_SUCCESS ) { 
      printf("Error en la lectura del bloque de memoria \n");
      return error; 
   } 
   temp = (char *)(vmeip - addr_in);
   *eip = (vm_address_t)(temp + (*addr)); 
   return KERN_SUCCESS; 
}

/* 
 * Funcion que reemplaza la pagina que contiene la direccion vmeip con una 
 * pagina local.
 * Escritura 
 * Amilcar Meneses Viveros 
 * [9-JULIO-2000]
 */ 
kern_return_t map_out(task_t *t, vm_address_t vmeip, vm_address_t addr)
{
   char addr_out; 
   kern_return_t error;

   addr_out = (char *)trunc_page(vmeip); 
   if ( (error = vm_write(*t , (vm_address_t) addr_out, (pointer_t) addr, 
        vm_page_size) )!= KERN_SUCCESS ) {
      printf("Salgo en funcion error \n");  
      return error; 
   }

   return KERN_SUCCESS;
}

/***************************************************************
 *        SIMULACION DE INSTRUCCIONES ESPECIALES               * 
 ***************************************************************/


/* 
 * Simula la instruccion pushf 
 */ 
void do_pushf (u_char *eip, i386_thread_state_t *state)
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 
    vmesp = (u_short *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

    /* Realiza simulacion en la pagina mapeada */ 
    *--esp = state->eflags; 
    state->esp -= sizeof(short);
    state->eip++; 

    /* Escribe a la maquina virtual el resultado de la simulacion */ 
    map_out(&new_t, (vm_address_t) vmesp, (vm_address_t) addr); 

    /* Borra la pagina copiada */ 
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
}

/* 
 * Simula la instruccion pushf para el registro de banderas de 16 bits 
 */ 
void do_pushf_32 (u_char *eip, i386_thread_state_t *state)
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 
    vmesp = (u_short *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

    /* Realiza simulacion en la pagina mapeada */ 
    *--esp = state->eflags; 
    state->esp -= sizeof(int);
    state->eip++; 

    /* Escribe a la maquina virtual el resultado de la simulacion */ 
    map_out(&new_t, (vm_address_t) vmesp, (vm_address_t) addr); 

    /* Borra la pagina copiada */ 
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
}


/* 
 * Simula la instruccion popf 
 */ 
void do_popf (u_char *eip, i386_thread_state_t *state) 
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 
    vmesp = (u_short *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

    state->eflags &= ~EFL_TSAFE; 
    state->eflags |= (*esp++ & EFL_TSAFE); 
    state->esp += sizeof(short);
    state->eip++; 

    /* Escribe a la maquina virtual el resultado de la simulacion  */ 
    /* No hay resultados que escribir...                           */
    /* map_out(&new_t, (vm_address_t) vmesp, (vm_address_t) addr); */ 
 

    /* Borra la pagina copiada */ 
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
}

/* 
 * Simula la instruccion popf para el registro de banderas de 16 bits
 */ 
void do_popf_32 (u_char *eip, i386_thread_state_t *state) 
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 
    vmesp = (u_short *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

    state->eflags &= ~EFL_TSAFE; 
    state->eflags |= (*esp++ & EFL_TSAFE); 
    state->esp += sizeof(int);
    state->eip++; 

    /* Escribe a la maquina virtual el resultado de la simulacion  */ 
    /* No hay resultados que escribir...                           */
    /* map_out(&new_t, (vm_address_t) vmesp, (vm_address_t) addr); */ 
 

    /* Borra la pagina copiada */ 
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
}

/* 
 * Simula la instruccion int 
 */
void do_int (u_char *eip, i386_thread_state_t *state) 
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 

//    vmesp = (u_char *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
//    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

/*
    *--esp = state->eflags; 
    *--esp = state->cs; 
    *--esp = state->eip; 
    state->esp -= 6;
*/
    /* Escribe a la maquina virtual el resultado de la simulacion 
     * (movimientos en la pila) 
     */ 
//    map_out(&new_t, (vm_address_t) vmesp, (vm_address_t) addr); 

    /* Borra la pagina copiada */ 
/*
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
 */ 
    /* Simulacion de las diferentes interrupciones (BIOS y DOS) */ 
    eip++; 
    switch(*eip) { 
    case BIOS_VIDEO: do_int10_bios(state); 
                     break; 
    }
    state->eip+=2; 
}

/* 
 * Simula la instruccion iret 
 */  
void do_iret (u_char *eip, i386_thread_state_t *state) 
{
    u_short  *vmesp, *esp; 
    char     *addr; 
    kern_return_t rtn; 

    /* Obtiene la direccion de la pila en la maquina virtual */ 
    vmesp = (u_char *)Address_8086(state->ss, state->esp); 

    /* Mapea la region de memoria a una pagina local */ 
    map_in(&new_t, vmesp, (vm_address_t *)&esp, (vm_address_t *)&addr);  

    state->eip = *esp++; 
    state->cs  = *esp++; 
    state->eflags &= ~EFL_SAFE; 
    state->eflags |= (*esp++ & EFL_SAFE);
    state->esp += 6;  

    /* Borra la pagina copiada */ 
    if ((rtn=vm_deallocate(task_self(),(vm_address_t) addr, PAGINA_VIRTUAL))
          !=KERN_SUCCESS ) { 
        mach_error("Fallo vm_deallocate ", rtn); 
        printf("Salgo en funcion do_pushf\n"); 
        return; 
    }
}

/* 
 * Simula la instruccion in 
 */  
void do_in (u_char *eip, i386_thread_state_t *state) 
{
    /* 
        La simulacion de la instruccion in depende de la simulacion de la 
        arquitectura completa del hardware donde se ejecuta el 8086/88 
     */ 
    return; 
}

/* 
 * Simula la instruccion out 
 */  
void do_out (u_char *eip, i386_thread_state_t *state) 
{
    /* 
        La simulacion de la instruccion out depende de la simulacion de la 
        arquitectura completa del hardware donde se ejecuta el 8086/88 
     */ 
    return; 
}

/*************************************************
 * Rutinas de simulacion y manejo de excepciones * 
 *************************************************/

void do_bad_instruction(u_char *eip, i386_thread_state_t *state)
{
   boolean_t	operand_size_override = FALSE;
   simpleMsg  mensaje; 
   init_msg(&mensaje); 

 /*  printf("manejo la simulacion de la instruccion: 0x%8.8x\n", *eip);*/

restart: 
   switch(*eip) {
   case I386_INT:
           if (stopException==YES) {
              mensaje.h.msg_id = EXCEPTION_MSG; 
              mensaje.body[0] =(char) *eip; 
              mensaje.body[1] = mensaje.body[2] = mensaje.body[3] = '\0'; 
           /*   printf("Mando el codigo %d 0x%8.8x", 
                      (unsigned int)mensaje.body[0], 
                                    mensaje.body[0]); */
              msg_send((msg_header_t *)&mensaje,MSG_OPTION_NONE,0);
              mutex_lock(ExcLock);
             while(!bandera)
                condition_wait(ExcUpdated,ExcLock);
              mutex_unlock(ExcLock);
              bandera = 0; 
           }
          /* printf("Fue INT\n"); */
           do_int(eip, state); 
           break;  
   case I386_OP_SIZE: 			/* Operador del i386 */ 
           operand_size_override = TRUE; 
           state->eip++; 
           *eip++; 
           goto restart; 
           break; 
   case I386_HLT: 
           printf("es HLT!!!\n");
           state->eip++; 
           break; 
   case I386_CLI:  
           printf("es CLI!!!\n");
           state->eip++; 
           break; 
   case I386_STI:  
           printf("es STI!!!\n");
           state->eip++; 
           break; 
   case I386_PUSHF: 
           if (stopException==YES) {
              mensaje.h.msg_id = EXCEPTION_MSG; 
              mensaje.body[0] =(char) *eip; 
              mensaje.body[1] = mensaje.body[2] = mensaje.body[3] = '\0'; 
              printf("Mando el codigo %d 0x%8.8x", 
                      (unsigned int)mensaje.body[0], 
                                    mensaje.body[0]); 
              msg_send((msg_header_t *)&mensaje,MSG_OPTION_NONE,0);
              mutex_lock(ExcLock);
             while(!bandera)
                condition_wait(ExcUpdated,ExcLock);
              mutex_unlock(ExcLock);
              bandera = 0; 
           }
           printf("Fue PUSHF\n"); 
           if (operand_size_override) 
                do_pushf_32(eip, state);
           else do_pushf(eip, state); 
     //    thread_terminate(vm_thread); 
     //    state->eip++;
     //    while(1) ; 
           break; 
   case I386_POPF: 
           printf("es POPF!!!\n"); 
           if (operand_size_override) 
                do_popf_32(eip, state);
           else do_popf(eip, state);
     //    thread_terminate(vm_thread);
     //    state->eip++; 
     //    while(1); 
           break; 
   case I386_LOCK:
           printf("es LOCK!!!\n");
           state->eip++;  
           break;  
   case I386_IRET: 
           printf("es IRET!!!\n");
           do_iret(eip, state); 
           break; 
   case I386_IN1:  
   case I386_IN2: 
   case I386_IN3: 
   case I386_IN4: 
           printf("es IN!!!\n");
           do_in(eip, state);
           break; 
   case I386_OUT1: 
   case I386_OUT2: 
   case I386_OUT3: 
   case I386_OUT4: break; 
   case I386_INS1:  
   case I386_INS2: break; 
   case I386_OUTS1:  
   case I386_OUTS2: break; 
   case I386_REPZ: break; 
   }

}

void do_breakpoint(u_char *eip, i386_thread_state_t *state)
{

}

void do_bad_access(u_char *eip, i386_thread_state_t *state, vm_offset_t addr)
{

}

void do_arithmetic(u_char *eip, i386_thread_state_t *state, int code, int subcode)
{

}


/* Rutina que maneja excepciones */ 
kern_return_t catch_exception_raise(port_t p, thread_t th, task_t tk, 
       int exception, int code, int subcode)
{
   int state_count; 
   i386_thread_state_t state; 
   kern_return_t error; 
   u_char *vmeip, *eip;  
   char *addr;

   state_count = i386_THREAD_STATE_COUNT; 
   if ((error=thread_get_state(th, i386_THREAD_STATE, 
      (thread_state_t)(&state), &state_count ))!=KERN_SUCCESS) { 
      mach_error("Fallo el llamado a thread_get_state ", error); 
      printf("Salgo de la funcion thread_V86\n"); 
      exit(-1); 
   }

   vmeip = (u_char *)Address_8086(state.cs,state.eip); 

   switch(exception) { 
   case EXC_SOFTWARE: // Finaliza el hilo o la tarea virtual 
        switch(code){ 
        case END_RUN_VM:
             error = thread_terminate(vm_thread);
             if (error != KERN_SUCCESS)
                printf("No se puede terminar la ejecucion, hilo no valido\n");
             else 
             printf("Se termina la ejecucion del hilo virtual\n");
             break;
        case END_TASK_VM: 
             error = task_terminate(new_t);
             if (error != KERN_SUCCESS)
                printf("No se puede destruir la tarea, tarea no valida\n");
             else 
             printf("Se destruyo la tarea virtual\n");
             break;  
        default: 
             printf("Otra excepcion de software\n");
             break;            
        }
        break; 
   case EXC_BAD_INSTRUCTION:
        map_in(&new_t, vmeip, (vm_address_t *)&eip, (vm_address_t *)&addr);  
   /*     printf("Mala instruccion\n");  */ 
 	do_bad_instruction (eip, &state); 
        if((error  = thread_set_state(th, i386_THREAD_STATE, 
            (thread_state_t)(&state), state_count))!=KERN_SUCCESS) { 
           mach_error("Fallo el llamado a thread_set_state", error); 
           printf("Salgo de la funcion thread_V86\n"); 
           exit(-1); 
        } 

        /* Borra la pagina copiada */ 
        if ((error=vm_deallocate(task_self(), (vm_address_t) addr, 
                              PAGINA_VIRTUAL)) != KERN_SUCCESS ) { 
           mach_error("Fallo vm_deallocate ", error); 
           printf("Salgo en funcion do_pushf\n"); 
           exit(-1); 
        }
	break;
   case EXC_BREAKPOINT:
	do_breakpoint (eip, &state);
        if((error  = thread_set_state(th, i386_THREAD_STATE, 
            (thread_state_t)(&state), state_count))!=KERN_SUCCESS) { 
           mach_error("Fallo el llamado a thread_set_state", error); 
           printf("Salgo de la funcion thread_V86\n"); 
           exit(-1); 
        }
	break;
   case EXC_BAD_ACCESS:
	do_bad_access(eip, &state, subcode);
        if((error  = thread_set_state(th, i386_THREAD_STATE, 
            (thread_state_t)(&state), state_count))!=KERN_SUCCESS) { 
           mach_error("Fallo el llamado a thread_set_state", error); 
           printf("Salgo de la funcion thread_V86\n"); 
           exit(-1); 
        }
        break;
   case EXC_ARITHMETIC: 
	do_arithmetic (eip, &state, code, subcode);
        if((error  = thread_set_state(th, i386_THREAD_STATE, 
            (thread_state_t)(&state), state_count))!=KERN_SUCCESS) { 
           mach_error("Fallo el llamado a thread_set_state", error); 
           printf("Salgo de la funcion thread_V86\n"); 
           exit(-1); 
        }
	break;
   default: 
        printf("Otra excepcion\n");
        break; 
   }


 /* printf("Procese interrupcion!!!!\n");  */

  return KERN_SUCCESS; 
}

void desp_ret_rcv(kern_return_t r)
{
   switch(r){ 
   case RCV_SUCCESS:           printf("RCV_SUCCESS\n"); break; 
   case RCV_INVALID_MEMORY:    printf("RCV_INVALID_MEMORY\n"); break; 
   case RCV_INVALID_PORT:      printf("RCV_INVALID_PORT\n"); break; 
   case RCV_TOO_LARGE:         printf("RCV_TOO_LARGE\n"); break; 
   case RCV_NOT_ENOUGH_MEMORY: printf("RCV_NOT_ENOUGH_MEMORY\n"); break; 
   case RCV_TIMED_OUT:         printf("RCV_TIMED_OUT\n"); break; 
   case RCV_INTERRUPTED:       printf("RCV_INTERRUPTED\n"); break; 
   case RCV_PORT_CHANGE:       printf("RCV_PORT_CHANGE\n"); break; 
   case END_RUN_VM:            printf("END_RUN_VM\n"); break; 
   default:                    printf("Otro mensaje! %d\n", r);
 } 
}

void process_exception(port_t pe)
{
   kern_return_t r; 

   struct msg { 
    msg_header_t   header; 
    char           data[64]; 
   } i_msg, o_msg, *i, *o; 

 /*  printf("[[[[[ 0x%8.8x %d %d ]]]]]\n", &port_exc, port_exc, pe); */

   while(1) {
     i = &i_msg; o = &o_msg; 
     i_msg.header.msg_size = 64; 
     i_msg.header.msg_local_port = pe;  
     r = msg_receive( &(i_msg.header), MACH_RCV_MSG, 0);  

 /*    printf("Entra mensaje!!\n");
      desp_ret_rcv(r); */
     if (r==RCV_SUCCESS) { 
        if (exc_server(&(i_msg.header), &(o_msg.header))) { 
          /* printf("Atiende la excepcion\n"); */
           r = msg_send( &(o_msg.header), MACH_SEND_MSG, 0); 
           if (r!=SEND_SUCCESS) {
              mach_error("exc_thread msg_receive ", r);
              exit(1); 
           } 
        } 
        else { 
           printf("exc_server no quiere el mensaje\n"); 
           exit(2); 
        }
     } 
     else { 
        mach_error("Error en <msg_receive> ",r);
        exit(3); 
     }  
   } 
}

void hilo()
{
   kern_return_t error; 
   int state_count; 
   i386_thread_state_t s; 
   char  *addr; 

   state_count = i386_THREAD_STATE_COUNT; 
   addr = (char *) PAGINA_VIRTUAL;

/*   printf("Entro en HILO!!\n"); */
   while (1) { 

     muestra_memoria(&new_t, (vm_address_t) addr, 256);

     if ((error=thread_get_state(vm_thread, i386_THREAD_STATE, 
        (thread_state_t)(&s), &state_count ))!=KERN_SUCCESS) { 
        mach_error("Fallo el llamado a thread_get_state ", error); 
        printf("Salgo de la funcion hilo\n"); 
        exit(-1); 
    }

    printf("eip=0x%8.8x  cs=0x%8.8x  ds=0x%8.8x  ss=0x%8.8x  es=0x%8.8x\n", 
            s.eip, s.cs, s.ds, s.ss, s.es); 
    printf("ebp=0x%8.8x eax=0x%8.8x ebx=0x%8.8x ecx=0x%8.8x edx=0x%8.8x\n", 
            s.ebp, s.eax, s.ebx, s.ecx, s.edx); 
    printf("edi=0x%8.8x esi=0x%8.8x esp=0x%8.8x\n\n", s.edi, s.esi, s.esp); 
    sleep(5); 
  } 
}


/* Funcion que crea una tarea que tendra el proceso monitor */ 
/* Amilcar Meneses Viveros */ 
/* 16-Julio-1998           */ 
void tarea_monitor() 
{ 
    char *dato;  

    tarea_virtual(&new_t); 
    printf("Crea Tarea Monitor\n"); 
    llena(&new_t, (vm_address_t *) &dato); 
}


/* Funcion para cargar el archivo "nombrearchivo" */ 
/* Amilcar Meneses Viveros */
/* 25-Enero-1999           */
int carga(const char *nombrearchivo)
{
   int ln; 
   char *dato; 
   dato = (char *) 0x00; 

   return carga_archivo(new_t, nombrearchivo,
                  (vm_address_t) dato, &ln); 
}

void maneja_hiloV86()
{ 
   task_t mt; 
   char   *dato; 

   mt = task_self(); 
   informa_tarea(&mt); 
   dato = (char *) 0x00;
   thread_V86(new_t, &vm_thread, (vm_address_t) dato, &port_exc); 
  /* printf("Antes de ejecutar!!! (%d)\n", port_exc); */
   informa_hilo(vm_thread); 

//   cthread_detach(cthread_fork((cthread_fn_t)hilo, NULL));
   thread_resume(vm_thread); 

   process_exception(port_exc); 
} 

/* Manejo de memoria a buffer */ 
/* 
 * Funcion que guarda en la cadena s, nbytes de la memoria de la tarea 
 * virtual (*t) a partir de la direccion addr. 
 * Amilcar Meneses Viveros 
 * [29-MARZO-1999] 
 */ 
void obten_memoria(task_t *t, vm_address_t addr, u_int nbytes, char *s)
{
   char  *datos, *addr_i, *addr_j, *addr_t, *addr_d; 
   int   nb, i, j, k, nl; 
   unsigned int d1, d2; 
   kern_return_t error; 
   char buffer[200]; 

   j = i = nl = 0; 
   addr_t = (char *)trunc_page(addr);
   if (error=vm_read(*t, (vm_address_t) addr_t, vm_page_size, 
      (pointer_t *)&datos, &nb) != KERN_SUCCESS ) { 
//      mach_error("vm_read fallo en obten memoria", error); 
      sprintf(s, "Error en la lectura del bloque de memoria \n");
      cthread_exit(0);
      return; 
   } 

   d1 = (unsigned int)addr; d2 = (unsigned int)addr_t; d1 -=d2;  
   addr_d = ((char *)addr) - addr_t; 
   addr_i =  datos + (u_int)addr_d; 

/* 
   printf("Estoy en obten_memoria"); 
   sprintf(buffer,"Datos de entrada:\n addr = [%8.8x]\n nbytes = %d\n", 
           addr, nbytes);  
   strcpy(s, buffer);
   sprintf(buffer, " datos = %8.8x\n Addr_i = %8.8x\n Addr_d = %8.8x\n", datos, addr_i, addr_d); 
   strcat(s, buffer);

   if ( (error=vm_deallocate(task_self(), (vm_address_t) datos, vm_page_size) ) 
         != KERN_SUCCESS ) {
//      mach_error("Fallo vm_deallocate ", error); 
      printf("Salgo de la funcion muestra_memoria\n"); 
      cthread_exit(0);
      exit(-1); 
   }

   return; 

   d1 = (long int)addr; d2 = (long int)addr_t; d1 -=d2;  
   addr_i =  datos; addr_i += d1; 
*/ 
   *s = '\0'; j=0;
   while(j<nbytes) { 
       sprintf(buffer, "[%4.4x]", addr+j); 
       strcat(s, buffer);
       addr_j = addr_i; 

       for (k=0; k<NBYTES; k++) {
           if (k==8) strcat(s, "- "); 
           sprintf(buffer, "%2.2x ", *addr_j++); 
           strcat(s, buffer); 
       }

       for (addr_j = addr_i, k=0; k < NBYTES; k++, addr_j++) {
           if (k == 7) strcat(s, "-");
           if (*(addr_j)>=' ' && *(addr_j)<='~') {  
              sprintf(buffer, "%c", *(addr_j)); 
              strcat(s, buffer); 
           }
           else  strcat(s, "."); 
       }

       strcat(s, "\n");
       addr_i += NBYTES; 
       j += NBYTES; 
   }

   if ( (error=vm_deallocate(task_self(), (vm_address_t) datos, vm_page_size) ) 
         != KERN_SUCCESS ) {
//      mach_error("Fallo vm_deallocate ", error); 
      printf("Salgo de la funcion muestra_memoria\n"); 
      cthread_exit(0); 
      return; 
   }
//   strcat(s, "\n"); 
}

void memoriaV86()
{
   simpleMsg  mensaje; 
   char  *addr, *addrL; 

   addr = (char *) PAGINA_VIRTUAL;
   addrL = addr + 256; 

   init_msg(&mensaje); 

   while(1) { 
//     printf("Muestro memoria\n"); 
     mensaje.h.msg_id = MEMORY_MSG; 

//     printf("HOLA HOLA HOLA\n");
     obten_memoria(&new_t, (vm_address_t) addr, 16, mensaje.body);
//     printf("Datos que lei:\n"); 
//     printf("[%s]", mensaje.body); 

//     strcpy(mensaje.body,"Hola\nCrayola"); 
     msg_send((msg_header_t *)&mensaje,MSG_OPTION_NONE,0);
     sleep(5); 
     if (addr<addrL) addr+=16; 
     else { 
        addr = (char *)PAGINA_VIRTUAL; 
        mensaje.h.msg_id = CLEAR_MSG; 
        msg_send((msg_header_t *)&mensaje,MSG_OPTION_NONE,0);
        sleep(5); 
     }
   }
}

/*******************************************************************/
/* FUNCION PRINCIPAL      					   */ 
/* Se crea un tarea hija, se le asigna el 1 er megabyte de memoria */ 
/* Se ejecuta una funcion para que escriba en esta localidad       */ 
/*******************************************************************/
void main_a()
{
   int ln; 
   char *dato; 
//   char *addr1, *addr; 
   task_t mt; 

   tarea_virtual(&new_t); 
   estado_mem8086(&new_t, LOAD_SEGMENT, LOAD_OFFSET);
   llena(&new_t, (vm_address_t *) &dato); 

   printf("Direccion Base <%8.8x>\n", (u_int) dato); 
   carga_archivo(new_t, "/Users/ameneses/tesis/VM/exc.com", 
               (vm_address_t) dato, &ln); 

   printf("\nBytes leidos: <%d>\n", ln);  
/*   muestra_memoria(&new_t,(vm_address_t) dato, PAGINA_VIRTUAL); 
   addr1 = (char *) PAGINA_VIRTUAL; 
    muestra_memoria(&new_t,(vm_address_t) addr1, 256);  */ 

     printf("\n Tarea Padre \n"); 
     mt = task_self(); 
     informa_tarea(&mt); 

   thread_V86(new_t, &vm_thread, (vm_address_t) dato, &port_exc); 
   printf("Antes de ejecutar!!! (%d)\n", port_exc); 
   informa_hilo(&vm_thread); 


   cthread_detach(cthread_fork((cthread_fn_t)hilo, NULL));
   thread_resume(vm_thread); 


// while(1) ; 

   process_exception(port_exc); 

   printf("Ejecuta!!!\n");
/*   addr = (char *) PAGINA_VIRTUAL; 
   while(1) { 
     printf("\n Hilo Virtual\n"); 
     informa_hilo(vm_thread); 
     printf("Muestra Memoria\n"); 
     muestra_memoria(&new_t, (vm_address_t) addr, 256);
     sleep(5); 
   } 
*/
/*   task_terminate(new_t); */ 

   printf("TERMINE!!!!!\n"); 

}